import scipy.constants as constants

h=constants.physical_constants['Planck constant in eV s'][0]
hbar=constants.physical_constants['Planck constant over 2 pi in eV s'][0]
c=constants.c
standardLength=1e-10
